#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>

#define MAX_TARGETS	32
static struct sockaddr_in targets[MAX_TARGETS];
static unsigned int targcnt = 0;

int main(int argc, char* argv[]) {
	const char* devname;
	unsigned int port;
	char* ep;
	int i;
	int master;
	int sock;
	unsigned int val;
	struct sockaddr_in sin;

	if (argc < 4) {
		fprintf(stderr, "usage:         vmpipe device localPort otherNode ...\n"
				"\n"
				"For example:   vmpipe /dev/vmnet2 20000 10.20.30.40:20000\n"
				"\n");
		return 20;
	}
	devname = argv[1];
	port = strtoul(argv[2], &ep, 0);
	if (ep && *ep) {
		fprintf(stderr, "Port number `%s' is not valid.\n", argv[2]);
		return 19;
	}
	if (port == 0) {
		port = 20000;
	}
	for (i = 3; i < argc; i++) {
		char* str = argv[i];
		char* pp;
		struct hostent* he;

		if (targcnt >= MAX_TARGETS) {
			fprintf(stderr, "Sorry, only %u remote nodes is possible\n", targcnt);
			return 16;
		}
		pp = strrchr(str, ':');
		if (pp) {
			*pp++ = 0;
			targets[targcnt].sin_port = htons(strtoul(pp, &ep, 0));
			if (targets[targcnt].sin_port == 0) {
				fprintf(stderr, "Invalid port number `%s' specified.\n", pp);
				return 18;
			}
		} else {
			targets[targcnt].sin_port = htons(port);
		}
		he = gethostbyname(str);
		if (!he) {
			fprintf(stderr, "Unknown host `%s': %s\n", str, hstrerror(h_errno));
			return 17;
		}
		memcpy(&targets[targcnt].sin_addr.s_addr, he->h_addr, 4);
		targets[targcnt].sin_family = AF_INET;
		targcnt++;
	}

	master = open(devname, O_RDWR);
	if (master == -1) {
		fprintf(stderr, "Cannot open `%s': %s\n", devname, strerror(errno));
		return 16;
	}
	fcntl(master, F_SETFL, O_NONBLOCK | fcntl(master, F_GETFL));
	val = IFF_UP|IFF_BROADCAST|IFF_PROMISC|IFF_MULTICAST|IFF_ALLMULTI|IFF_RUNNING;
	ioctl(master, SIOCSIFFLAGS, &val);

	sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock == -1) {
		fprintf(stderr, "Cannot create socket: %s\n", strerror(errno));
		close(master);
		return 15;
	}
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	sin.sin_port = htons(port);
	if (bind(sock, (struct sockaddr*)&sin, sizeof(sin))) {
		fprintf(stderr, "Cannot bind to socket: %s\n", strerror(errno));
		close(sock);
		close(master);
		return 14;
	}
	fcntl(sock, F_SETFL, O_NONBLOCK | fcntl(sock, F_GETFL));
	printf("Processing packets...\n");
	while (1) {
		fd_set inp;
		int mx;
		int c;

		FD_ZERO(&inp);
		FD_SET(sock, &inp);
		FD_SET(master, &inp);
		mx = sock;
		if (master > mx) {
			mx = master;
		}
		c = select(mx + 1, &inp, NULL, NULL, NULL);
		if (c == -1) {
			if (errno == EINTR) {
				continue;
			}
			fprintf(stderr, "Something wicked happened during wait: %s\n", strerror(errno));
			close(sock);
			close(master);
			return 13;
		}
		if (c == 0) {
			continue;
		}
		if (FD_ISSET(sock, &inp)) {
			socklen_t sln;
			ssize_t ln;
			unsigned char data[2000];

			sln = sizeof(sin);
			ln = recvfrom(sock, data, sizeof(data), MSG_TRUNC, (struct sockaddr*)&sin, &sln);
			if (ln >= 0) {
				/* Verify sin... */
				write(master, data, ln);
			}
		}
		if (FD_ISSET(master, &inp)) {
			unsigned char data[2000];
			ssize_t ln;

			ln = read(master, data, sizeof(data));
			if (ln >= 0) {
				unsigned int idx;

				for (idx = 0; idx < targcnt; idx++) {
					sendto(sock, data, ln, 0, (struct sockaddr*)&targets[idx], sizeof(targets[idx]));
				}
			}
		}
	}
}


