#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#ifdef _WIN32
#include <io.h>
#include <winsock2.h>

typedef int socklen_t;
typedef int ssize_t;
#define MSG_TRUNC 0
#define fcntl(s, k, v) (0)
#define SIOCSIFFLAGS CTL_CODE(0x8102, 0x804, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define VNETBIND  CTL_CODE(0x8102, 0x811, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define VNETRECV  CTL_CODE(0x8102, 0x80B, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IFF_UP        0x001
#define IFF_BROADCAST 0x002
#define IFF_PROMISC   0x008
#define IFF_MULTICAST 0x010
#define IFF_RUNNING   0x020
#define IFF_ALLMULTI  0x200
#else
#include <netdb.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>

#define CloseHandle(x) close(x)

#endif

#define MAX_TARGETS	32
static struct sockaddr_in targets[MAX_TARGETS];
static unsigned int targcnt = 0;
static unsigned int debug = 0;

OVERLAPPED sm;

static const char* mystrerror(int errorcode) {
	static char buf[4096];

	if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, 0, errorcode, 0, buf, sizeof(buf), NULL) == 0) {
		return "Unknown error";
	}
	return buf;
}

void processReceive(HANDLE master,
		    void* data,
		    DWORD len,
		    struct sockaddr_in* sa,
		    INT salen) {
	if (debug) {
		fprintf(stderr, "Received %u bytes from network\n", len);
	}
	WriteFile(master, data, len, &len, NULL);
}

void processRead(SOCKET sock,
		 void* data,
		 DWORD len) {
	unsigned int i;

	if (debug) {
		fprintf(stderr, "Sending %u bytes to network\n", len);
	}
	for (i = 0; i < targcnt; i++) {
		WSABUF sb;
		DWORD e;

		sb.buf = data;
		sb.len = len;
		if (WSASendTo(sock, &sb, 1, &e, 0, (struct sockaddr*)&targets[i], sizeof(targets[i]), &sm, NULL) == SOCKET_ERROR) {
			if (WSAGetLastError() == WSA_IO_PENDING) {
				WSAGetOverlappedResult(sock, &sm, &e, TRUE, &e);
			}
		}
	}
}

int initSockets(void) {
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;

	wVersionRequested = MAKEWORD( 2, 0 );

	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		fprintf(stderr, "Unable to initialize WinSock\n");
		return 21;
	}
	if (LOBYTE( wsaData.wVersion ) != 2 ||
	    HIBYTE( wsaData.wVersion ) != 0 ) {
		fprintf(stderr, "Sorry, this version 2.0 of WinSock is required\n");
		WSACleanup();
		return 22;
	}
	return 0;
}

int main(int argc, char* argv[]) {
	const char* devname;
	unsigned int port;
	char* ep;
	int i;
#ifdef _WIN32
	HANDLE master;
	SOCKET sock;
	WSAOVERLAPPED socketO;
	int socketR;
	unsigned char masterB[2000];
	unsigned char socketB[2000];
	DWORD masterL;
	DWORD socketL;
	struct sockaddr_in socketRI;
	INT socketRIL;
	DWORD socketF;
	HANDLE masterE;
#else
	int master;
	int sock;
#endif
	unsigned int val;
	struct sockaddr_in sin;

	if (initSockets()) {
		return 21;
	}
	if (argc > 1 && !strcmp(argv[1], "-d")) {
		debug = 1;
		argc--;
		argv++;
	}
	if (argc < 4) {
		fprintf(stderr, "usage:         vmpipe device localPort otherNode ...\n"
				"\n"
				"For example:   vmpipe vmnet2 20000 10.20.30.40:20000\n"
				"\n");
		return 20;
	}
	devname = argv[1];
	port = strtoul(argv[2], &ep, 0);
	if (ep && *ep) {
		fprintf(stderr, "Port number `%s' is not valid.\n", argv[2]);
		return 19;
	}
	if (port == 0) {
		port = 20000;
	}
	for (i = 3; i < argc; i++) {
		char* str = argv[i];
		char* pp;
		struct hostent* he;

		if (targcnt >= MAX_TARGETS) {
			fprintf(stderr, "Sorry, only %u remote nodes is possible\n", targcnt);
			return 16;
		}
		pp = strrchr(str, ':');
		if (pp) {
			*pp++ = 0;
			targets[targcnt].sin_port = htons(strtoul(pp, &ep, 0));
			if (targets[targcnt].sin_port == 0) {
				fprintf(stderr, "Invalid port number `%s' specified.\n", pp);
				return 18;
			}
		} else {
			targets[targcnt].sin_port = htons(port);
		}
		he = gethostbyname(str);
		if (!he) {
			fprintf(stderr, "Unknown host `%s': %s\n", str, mystrerror(WSAGetLastError()));
			return 17;
		}
		memcpy(&targets[targcnt].sin_addr.s_addr, he->h_addr, 4);
		targets[targcnt].sin_family = AF_INET;
		targcnt++;
	}

	if (strncmpi(devname, "vmnet", 5)) {
		master = CreateFile(devname, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (master == INVALID_HANDLE_VALUE) {
			fprintf(stderr, "Cannot open `%s': %s\n", devname, mystrerror(GetLastError()));
			return 16;
		}
	} else {
		master = CreateFile("\\\\.\\VMnetUserif", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (master == INVALID_HANDLE_VALUE) {
			fprintf(stderr, "Cannot open `%s': %s\n", devname, mystrerror(GetLastError()));
			return 16;
		}
		{
			DWORD rtn;

			val = strtoul(devname + 5, NULL, 0);
			if (!DeviceIoControl(master, VNETBIND, &val, sizeof(val), NULL, 0, &rtn, &sm)) {
				if (GetLastError() == ERROR_IO_PENDING) {
					GetOverlappedResult(master, &sm, &rtn, TRUE);
				}
			}
		}
	}
	{
		DWORD rtn;

		val = IFF_UP|IFF_BROADCAST|IFF_PROMISC|IFF_MULTICAST|IFF_ALLMULTI|IFF_RUNNING;
		if (!DeviceIoControl(master, SIOCSIFFLAGS, &val, sizeof(val), NULL, 0, &rtn, &sm)) {
			if (GetLastError() == ERROR_IO_PENDING) {
				GetOverlappedResult(master, &sm, &rtn, TRUE);
			}
		}
	}
	sock = WSASocket(PF_INET, SOCK_DGRAM, IPPROTO_UDP, NULL, 0, WSA_FLAG_OVERLAPPED);
	if (sock == INVALID_SOCKET) {
		fprintf(stderr, "Cannot create socket: %s\n", mystrerror(WSAGetLastError()));
		CloseHandle(master);
		return 15;
	}
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	sin.sin_port = htons(port);
	if (bind(sock, (struct sockaddr*)&sin, sizeof(sin))) {
		fprintf(stderr, "Cannot bind to socket: %s\n", mystrerror(WSAGetLastError()));
		close(sock);
		CloseHandle(master);
		return 14;
	}
	printf("Processing packets...\n");

	memset(&socketO, 0, sizeof(socketO));
	memset(&sm, 0, sizeof(sm));
	masterE = WSACreateEvent();
	socketO.hEvent = WSACreateEvent();
	sm.hEvent = WSACreateEvent();
	socketR = 0;
	DeviceIoControl(master, VNETRECV, &masterE, sizeof(masterE), NULL, 0, &masterL, NULL);
	while (1) {
		if (WaitForSingleObject(masterE, 0) == WAIT_OBJECT_0) {
			if (ReadFile(master, masterB, sizeof(masterB), &masterL, NULL)) {
				processRead(sock, masterB, masterL);
			} else {
				fprintf(stderr, "Unexpected error returned by ReadFile: %s\n", mystrerror(GetLastError()));
			}
		}
		if (socketR == 1) {
			if (WSAGetOverlappedResult(sock, &socketO, &socketL, FALSE, &socketF)) {
				processReceive(master, socketB, socketL, &socketRI, socketRIL);
				socketR = 0;
			} else if (WSAGetLastError() == WSA_IO_INCOMPLETE) {
				;
			} else {
				fprintf(stderr, "Unexpected error returned by WSAGetOverlappedResult: %s\n", mystrerror(WSAGetLastError()));
				socketR = 0;
			}
		}
		if (socketR == 0) {
			WSABUF rb;

			socketF = 0;
			socketRIL = sizeof(socketRI);
			rb.buf = socketB;
			rb.len = sizeof(socketB);
			if (WSARecvFrom(sock, &rb, 1, &socketL, &socketF, (struct sockaddr*)&socketRI, &socketRIL, &socketO, NULL) != SOCKET_ERROR) {
				processReceive(master, socketB, socketL, &socketRI, socketRIL);
			} else if (WSAGetLastError() == WSA_IO_PENDING) {
				socketR = 1;
			} else {
				fprintf(stderr, "Unexpected error returned by WSARecvFrom: %s\n", mystrerror(WSAGetLastError()));
			}
		}
		if (socketR == 1) {
			WSAEVENT ev[2];

			ev[0] = masterE;
			ev[1] = socketO.hEvent;
			WSAWaitForMultipleEvents(2, ev, FALSE, WSA_INFINITE, TRUE);
		}
	}
}


